# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Contains the get_metric function."""

from typing import Union
from dlfs.metrics import Metric, Accuracy
from dlfs.losses import *


def get_metric(name: str) -> Union[Metric, LossFunction]:
    """Returns a metric object by name.
    """

    if name == 'accuracy':
        return Accuracy()
    elif name == 'mae':
        return MAE()
    elif name == 'mse':
        return MSE()
    else:
        raise ValueError(f'Unknown metric: {name}')
